from typing import Literal
import logging

# ==========================================================================================
#                                       LOGGER
# ==========================================================================================
_logger: logging.Logger | None = None

LogLevel = Literal["debug", "info", "warning", "error", "critical"]
_LOG_LEVELS = ("debug", "info", "warning", "error", "critical") # For runetime use
_base_log_level = "info"

def set_logger(logger: logging.Logger) -> None:
    """Sets a global logger to log all the function print to it

    Args:
        logger (logging.Logger): Logger
    """
    global _logger
    _logger = logger

def get_logger() -> logging.Logger | None:
    """Get the logger instance used in the module

    Returns:
        logging.Logger | None: Logger
    """
    return _logger


def set_base_log_level(base_log_level: LogLevel):
    """Sets the base log level for 'print_log' without spcifying the log level

    Args:
        base_log_level (LogLevel): Log level of logger from logging

    Raises:
        ValueError: If the log level is not a valid one.
    """
    global _base_log_level
    if base_log_level not in _LOG_LEVELS:
        raise ValueError(f"The log level '{base_log_level}' is not in those listed: {_LOG_LEVELS}")
    _base_log_level = base_log_level

def get_base_log_level() -> LogLevel:
    """Get the Module base log levelv

    Returns:
        LogLevel: Module base log levelv
    """
    return _base_log_level

# ==========================================================================================
#                                       Verbose
# ==========================================================================================
_verbose: bool = True
def set_verbose(verbose: bool) -> None:
    """Sets the module verbose to true or false. If false none of the functions that print debug
    lines will print. Yet, the print_... functions like print_separator or print_log will still print.
    Yet, this value can be ignored when calling some functions with parameter verbose.

    Args:
        verbose (bool): Verbose or not
    """
    global _verbose
    _verbose = verbose

def get_verbose(verbose: bool = None) -> bool:
    """Returns general verbose if one is not passed. Else 
    follows function passed verbose. 

    Args:
        verbose (bool, optional): _description_. Defaults to None.

    Returns:
        bool: _description_
    """
    return verbose if verbose is not None else _verbose


import os
import json
import shutil
from typing import Any, Callable
from pathlib import Path
from natsort import natsorted

from .print_utils import print_status, print_warn, print_error, print_log
from .config import get_verbose
# ==========================================================================================
#                                       JSON 
# ==========================================================================================
def save_json(save_path: str, content: Any, verbose: bool = None):
    """
    Saves a Python object as a JSON file.

    Args:
        save_path (str): Full path including file name to save the JSON.
        content (Any): The content to save (must be JSON-serializable).
        verbose (bool): Wether to print or not. If not set, will follow general config.

    Returns:
        Any: The original content.
    """
    if get_verbose(verbose):
        print_log(f"Saving output at {save_path}...")
    
    # Extra safety for empty paths
    dir_path = os.path.dirname(save_path)
    if dir_path:
        os.makedirs(dir_path, exist_ok=True)

    with open(save_path, "w", encoding="utf-8") as out_json:
        json.dump(content, out_json, indent=4)
    
    return content


def load_json(save_path: str, verbose: bool = None) -> Any:
    """
    Loads JSON content from a file. Returns an empty dict if the file does not exist.

    Args:
        save_path (str): Full path including file name to read the JSON from.
        verbose (bool): Wether to print or not. If not set, will follow general config.

    Returns:
        Any: The loaded content, or an empty dict if the file does not exist.
    """
    if not os.path.exists(save_path):
        if get_verbose(verbose):
            print_warn(f"NO FILE AT {save_path}. Returning empty dict...")
        return dict()

    if get_verbose(verbose):
        print_log(f"Loading output from {save_path}...")

    with open(save_path, "r", encoding="utf-8") as out_json:
        content = json.load(out_json)
    
    return content


# ==========================================================================================
#                                       DIRECTORIES
# ==========================================================================================
def check_dirs_existance(dir_or_dirs: str | list[str], raise_error: bool = True) -> list[str]:
    """
    Checks if all specified dir_or_dirs exist.

    Args:
        dir_or_dirs (str | list[str]): Directory or list of directory paths to check.
        raise_error (bool): If True, raises a KeyError if any directory does not exist.
                            If False, the function will return false instead of raising an error.

    Raises:
        KeyError: If any of the dir_or_dirs do not exist.
    
    Returns:
        list[str]: A list of missing directories, or an empty list if all exist.
    """
    if isinstance(dir_or_dirs, str):
        dir_or_dirs = [dir_or_dirs]

    missing = [d for d in dir_or_dirs if not os.path.exists(d)]
    if missing and raise_error:
        raise KeyError(f"Some paths were not found: {missing}")
    
    return missing


def make_dirs(dir_or_dirs: str | list[str]) -> list[bool]:
    """
    Creates the specified dir_or_dirs if they do not already exist.

    Args:
        dir_or_dirs (str | list[str]): Directory or list of directory paths to create.
    
    Returns:
        list[bool]: for each directory 'created', whether or not it existed before.
    """
    if isinstance(dir_or_dirs, str):
        dir_or_dirs = [dir_or_dirs]

    existed = []
    for directory in dir_or_dirs:
        existed.append(os.path.exists(directory))
        os.makedirs(directory, exist_ok=True)

    return existed

def clear_directories(dir_or_dirs: str | list[str], remove_folder: bool = False, verbose: bool = None):
    """Clears out each directory in `dir_or_dirs`. 
    If dir_or_dirs is just a directory (string) just clear that one. Else
    loop over all the directories and clean them all.
    If `remove_folder` is True, the entire folder is removed; otherwise only its contents are deleted.

    Args:
        dir_or_dirs (str | list[str]): Path/s to directories.
        remove_folder (bool): If True, delete the folder itself.
        verbose (bool): Wether to print or not. If not set, will follow general config.

    Notes:
        - If the folder does not exist or is not a directory, a message is printed.
        - This function is intended for temporary folders or cleanup tasks.
    """
    if isinstance(dir_or_dirs, str): 
        dir_or_dirs = [dir_or_dirs]

    for d in dir_or_dirs:
        p = Path(d)
        if not p.exists() or not p.is_dir():
            if get_verbose(verbose):
                print_warn(f"{d!r} not found or not a dir")
            continue

        if remove_folder:
            shutil.rmtree(p)
        else:
            for child in p.iterdir():
                (shutil.rmtree(child) if child.is_dir() else child.unlink())

        if get_verbose(verbose):
            print_log(f"{'Removed' if remove_folder else 'Cleared'} {d}")



# ==========================================================================================
#                                       LIST DIR
# ==========================================================================================
def list_dir_files(path: str, max_files: int = None, recursive: bool = False, nat_sorting: bool = True, absolute_path: bool = True) -> tuple[list[str], int]:
    """Given a path and (optionally) the number of max files, loads a sorted list of
    all the files that are found in that folder (with at most max_files if passed).
    Then returns tha list of files and the number of those that have been loaded.
    If max_files is less than the actual number of files, it's okey. 
    By default it sorts the files as the os / humans would do. If you need python sorting
    use nat_sorting = False

    Args:
        path (str): Path to the folder
        max_files (int, optional): Max number of files to load. Defaults to None.
        recursive (bool, optional): If false just return the root level files in the path. If true list ALL files in any sub folder. Default to False.
        nat_sorting (bool, optional): Whether or not to sort the files as the os (naturally humans) do. Defualt to True.
        absolute_path (bool, optional): Whether or not return the absolute path (path + file names) or just the file names from path (files names). Defualt to True.
            absolute_path = True: 
                >>> path = "./test" -> ["./test/file1","./test/file2","./test/folder/file3"]
            absolute_path = False: 
                >>> path = "./test" -> ["/file1","/file2","/folder/file3"]


    Returns:
        tuple[list[str], int]: The list of file names and the number of those that have been listed
    """
    if not os.path.isdir(path):
        print_error(f"NO SUCH DIRECTORY: {path!r}")
        return [], 0
    
    # ========= GET ALL THE FILES =========
    if not recursive:
        aux = os.listdir(path)
        dir_list = [
            os.path.join(path, entry)
            for entry in aux
            if os.path.isfile(os.path.join(path, entry))
        ]
            
    else: # Loop over each subfile and add the whole path
        dir_list = []
        for dirpath, _, filenames in os.walk(path):
            for f in filenames:
                full_path = os.path.join(dirpath, f)
                dir_list.append(full_path)

    # ========= SORTHING =========
    if nat_sorting:
        dir_list = natsorted(dir_list)
    else:
        dir_list = list(sorted(dir_list))

    # ========= REMOVE ABS PATH =========
    if not absolute_path:
        dir_list = [os.path.relpath(f, path) for f in dir_list]


    # ========= CUT THE LIST =========
    dir_list = dir_list[:max_files]
    n_files = len(dir_list)

    return dir_list, n_files


def change_file_ext(path: str, new_extension: str) -> str:
    """
    Changes the extension of a given file path.

    Args:
        path (str): The original file path.
        new_extension (str): The new file extension (with or without a leading dot: '.txt' or 'txt').

    Returns:
        str: The file path with the new extension.
    """
    if not new_extension.startswith("."):
        new_extension = "." + new_extension
        
    return str(Path(path).with_suffix(new_extension))


def manage_temp_files(call_key: str, temp_folder_path: str, function: Callable, _verbose: bool = None, keep_logs: bool = True, *args, **kwargs) -> Any:
    """Manages temporal files for saving certain function outputs.

    When called for the first time, it will execute the function and save the file
    into a temporal folder. If by any reason (crash) this is called again with the
    same key, instead of executing process again, it will load the previous saved output

    Args:
        call_key (str): Unique as possible key for the call
        temp_folder_path (str): path to the temporal files folder
        to_execute (Callable): Function / process to be executed. Will pass all extra parameters
        *args: Positional arguments to pass to the function
        **kwargs: Keyword arguments to pass to the function

    Returns:
        Any: Whatever the process returns

    Examples:
    >>> for i in range(10):
    >>>     manage_temp_files(
    >>>         i,                     # call_key
    >>>         "./temp",              # temp_folder_path
    >>>         document_to_llm,       # function
    >>>         llm_client,            # *args -> function arg 1
    >>>         pages[i],              # *args -> function arg 2
    >>>         verbose=False          # **kwargs -> function keyword arg
    >>>     )
    """
    temp_path = os.path.join(temp_folder_path, f"{call_key}.json")

    if os.path.exists(temp_path):
        if get_verbose(_verbose):
            print_log(f" - Skiping '{call_key}': Temp file found at '{temp_path}'")
        return load_json(temp_path, verbose=False)

    if get_verbose(_verbose):
        if keep_logs:
            print_log(f" - Executing '{function.__name__}' for '{call_key}'...")
        else:
            print_status(f" - Executing '{function.__name__}' for '{call_key}'...")
    result = function(*args, **kwargs)

    save_json(temp_path, result, verbose=False)
    return result




import dataclasses
from argparse import Namespace

# =================================================
#                   HELPER FUNCTIONS
# =================================================
def parse_seconds_to_minutes(sec: float) -> str:
    """
    Parses a duration in seconds into a formatted string with hours, minutes, and seconds.

    Args:
        sec (float): Number of seconds.

    Returns:
        str: Formatted time string (e.g., '01 hrs, 05 mins, 30.1234 sec').
    """
    hours = int(sec // 3600)
    minutes = int((sec % 3600) // 60)
    seconds = int(sec % 60)
    decimals = int((sec % 1) * 10000)

    if hours > 0:
        return f"{hours:02} hrs, {minutes:02} mins, {seconds:02}.{decimals:04} sec"
    elif minutes > 0:
        return f"{minutes:02} mins, {seconds:02}.{decimals:04} sec"
    else:
        return f"{seconds:02}.{decimals:04} sec"


# =================================================
#                   DATACASS FUNCTIONS
# =================================================
def args_to_dataclass(args: Namespace, data_class: type):
    """From the args namespace, fillup a dataclass.

    It will change all the fields that have ben added to the args.
    If a field is not added in the args will be ignored.
    Fields in the args that are not in the Config this will be ignored.

    Args:
        args (Namespace): Parsed arguments. 
        data_class (type): Dataclass or any dataclass type from dataclasses.

    Returns:
        Configuration: Configuration with args values.
    """
    fields = {f.name for f in dataclasses.fields(data_class)}
    filtered = {k: v for k, v in vars(args).items() if k in fields and v is not None}
    return data_class(**filtered)


"""Module with all the Time traking utils functions."""

import time
import json
from typing import Tuple, TextIO

from .print_utils import print_separator, print_warn, print_time
from .other_utils import parse_seconds_to_minutes
# =================================================
#                   TIME TRACKER
# =================================================
        
class TimeTracker:
    """
    Class for tracking the time of a process. It allows to track the time of different points in the process and save them in a json file.
    It also allows to track the time of different laps in the process.
    """
    def __init__(self, name: str, start_track_now: bool = False ):
        self.name = name
        self.hist: dict[str, Tuple[float, float]] = dict()
        self.started: bool = False
        self.last_time: float = -1.0
        
        self.lap_hist = dict()
        self.lap_runing: bool = False
        self.lap_number: int = 0
        
        if start_track_now:
            self.start(verbose=False)
            
        print_separator(f"⏳ TIME TRACKER '{name}' INITIALIZED{'. STARTING NOW' if start_track_now else ''}! ⏳")
    
    def start(self, verbose: bool = True, space: bool = True):
        """
        Starts traking the time
        """
        self.started = True
        self.track("START", verbose=verbose, space=space)
        
    
    def track(self, tag: str, verbose: bool = True, space: bool = False, mute_warning: bool = False) -> float:
        """
        Track the time of a certain point and add it a tag. Return time since las track
        """
        if not self.started and not mute_warning:
            print_warn("WARNING: Traking without startng, will call start.")
            self.start()
            
        t = time.time()
        diff = t - self.last_time if self.last_time > 0 else 0
        
        if self.lap_runing:
            if tag in self.lap_hist:
                tag = f"{tag}_{self.lap_number}"
            self.lap_hist[tag] = (t, diff)
        else:
            if tag in self.hist:
                tag = f"{tag}_"
            self.hist[tag] = (t, diff)
        
        if verbose: 
            print_tag = tag if not self.lap_runing else f"{tag} lap {self.lap_number}"
            print_time(diff, prefix=f"⏳ {print_tag}", sufix=" ⏳", space=space)
            
        self.last_time = t
        return diff
    
    # ============================================================================
    #                              LAPS MANAGEMENT
    # ============================================================================
    def start_lap(self, N: int = None, verbose: bool = False, mute_warning: bool = False) -> int:
        """Starts a new lap with its oun metrics and returns the number of the current started lap

        Args:
            N (int, optional): Total number of potential laps. Defaults to None.
            verbose (bool, optional): Print the number of the lap. Defaults to False.
            mute_warning (bool, optional): Show or not warnings. Defaults to False.

        Returns:
            int: Number of the current started lap.
        """
        self.lap_runing = True
        self.lap_number += 1
        if len(self.lap_hist) > 0 and not mute_warning:
            print_warn("WARNING: Starting lap without finishing previous. The records will be overritten.")
            
        t = time.time()    
        self.lap_hist["START_LAP"] = (t, 0)
        self.last_time = t
        
        if verbose:
            print(f"⏳ Starting lap num {self.lap_number}{f'/{N} ' if N is not None else ''}⏳!")
            
        return self.lap_number
        
    def finish_lap(self):
        """Finish lap and add the point trak to the list
        """
        self.lap_runing = False
        
        t = time.time()
        self.lap_hist["FINISH_LAP"] = (t, t-self.lap_hist["START_LAP"][0])
        
        # Update possible previous times
        for tag, (t, diff) in self.lap_hist.items():
            if tag in self.hist:
                _, prev_diff = self.hist[tag]
                self.hist[tag] = (t, prev_diff + diff)
            else:
                self.hist[tag] = (t, diff)
                
                
        self.lap_hist = dict()
        
    # ============================================================================
    #                              STIMATE TIME
    # ============================================================================
    def stimate_lap_time(self, N: int, mute_warning: bool = False):
        """Stimate the time to finish N laps after all the alread finished laps

        Args:
            N (int): Total number of laps (including those already done)
            mute_warning (bool, optional): Whether or not mute the waning about not running laps. Defaults to False.
        """
        if not self.lap_runing and not mute_warning:
            print_warn("WARNING: Stimating lap without starting it. Returning...")
            return
            
        t_f_end = time.time()
        eta = (N - self.lap_number) * (t_f_end - self.hist["START"][0]) / self.lap_number
        print_time(
            t_f_end - self.lap_hist["START_LAP"][0], 
            prefix="Total ", 
            sufix=f". ETA: {parse_seconds_to_minutes(eta)}"
        )
    
    # ============================================================================
    #                              METRICS MANAGEMENT
    # ============================================================================
        
    def get_metrics(self, n: int = None, initial_tag: str = "START") -> dict:
        """
        Return a dict with all the metrics with the form: tag: (time, diff) 
        Added Normalized if n of samples is passed with the form: tag: (time, diff, diff/n) 
        
        initial_tag change it in case it hasn't been set as 'START' for the first track
        """
        t = time.time()
        if len(self.hist) > 0: 
            if initial_tag not in self.hist:
                print_warn(f"WARNING: Passed initial tag '{initial_tag}' not found in history. Setting to first.")
                initial_tag = next(iter(self.hist)) # Getting the firts added tag
            self.hist["TOTAL"] = (t, t - self.hist[initial_tag][0])
            
        else:
            print_warn("WARNING: Getting metrics with 0 tracked points. This will return an empty dict.")
        
        if n is not None:
            res_hist =  {
                tag: (time, diff, diff/n) for tag, (time, diff) in self.hist.items()
            }
        else:
            res_hist = self.hist.copy()
        
        if "START_LAP" in res_hist:
            res_hist.pop("START_LAP")
        return res_hist
        
    def save_metric(self, save_path: str, n: int = None) -> dict:
        """Compute metrics, save them into a file and return them 

        Args:
            save_path (str): Save path for the metrics
            n (int, optional): 'Number of files' processed to get and avg. Defaults to None.

        Returns:
            dict: Computed metrics
        """
        metrics = self.get_metrics(n)
        
        with open(save_path, "w") as f:
            json.dump(metrics, f)
            
        return metrics
        
    def print_metrics(self, n: int = None, out_file: TextIO = None) -> dict:
        """Compute and print the metrics. Optionally into a file.

        Args:
            n (int, optional): 'Number of files' processed to get and avg. Defaults to None.
            out_file (TextIO, optional): File where printing should be done. Defaults to None.

        Returns:
            dict: Computed metrics
        """
        metrics = self.get_metrics(n)
        metrics.pop('START', None)
        print("")
        if n is not None:
            print(f"Processed {n} files in total\n", file=out_file)
        
        for tag, records in metrics.items():
            diff = records[1]
            
            print_time(diff, n_files=n, prefix=tag, out_file=out_file)
                    
        return metrics
        



import logging
from typing import Literal, Optional, TextIO


from .other_utils import parse_seconds_to_minutes
from .config import get_logger, get_base_log_level, LogLevel

# ==========================================================================================
#                                       LOGGER
# ==========================================================================================

def print_log(
    text: str, 
    end: str = "\n", 
    logger: logging.Logger = None,
    log_level: LogLevel = None,
    out_file: TextIO = None
) -> None:
    if logger is None:
        logger = get_logger()
    if log_level is None:
        log_level = get_base_log_level()

    if logger: # Make sense bc get_logger may return None
        # get he correct level and if not listed use info
        log_func = getattr(logger, log_level, logger.info)
        log_func(text)
    elif out_file:
        print(text, file=out_file, end=end)
    else:
        print(text, end=end)


# ==========================================================================================
#                                       GENERAL
# ==========================================================================================
_separators_max_length = 128
_separators = {
    "short" : "_"*int(_separators_max_length/4),
    "normal": "_"*int(_separators_max_length/2),
    "long"  : "_"*int(_separators_max_length),
    "super" : "="*int(_separators_max_length),
    "start" : "="*int(_separators_max_length),
}
SepType = Literal["SHORT", "NORMAL", "LONG", "SUPER", "START"]

_colors = {
    # Regular colors
    "black": "\033[0;30m",
    "red": "\033[0;31m",
    "green": "\033[0;32m",
    "yellow": "\033[0;33m",
    "blue": "\033[0;34m",
    "purple": "\033[0;35m",
    "cyan": "\033[0;36m",
    "white": "\033[0;37m",

    # Other styles
    "reset": "\033[0m",
    "italic": "\033[3m",
    "bold_italic": "\033[1;3m",
    "underline": "\033[4m",
    "strikethrough": "\033[9m",

    # Bold
    "bold_black": "\033[1;30m",
    "bold_red": "\033[1;31m",
    "bold_green": "\033[1;32m",
    "bold_yellow": "\033[1;33m",
    "bold_blue": "\033[1;34m",
    "bold_purple": "\033[1;35m",
    "bold_cyan": "\033[1;36m",
    "bold_white": "\033[1;37m",

    # Underline
    "underline_black": "\033[4;30m",
    "underline_red": "\033[4;31m",
    "underline_green": "\033[4;32m",
    "underline_yellow": "\033[4;33m",
    "underline_blue": "\033[4;34m",
    "underline_purple": "\033[4;35m",
    "underline_cyan": "\033[4;36m",
    "underline_white": "\033[4;37m",

    # Background
    "bg_black": "\033[40m",
    "bg_red": "\033[41m",
    "bg_green": "\033[42m",
    "bg_yellow": "\033[43m",
    "bg_blue": "\033[44m",
    "bg_purple": "\033[45m",
    "bg_cyan": "\033[46m",
    "bg_white": "\033[47m",

    # High Intensity
    "hi_black": "\033[0;90m",
    "hi_red": "\033[0;91m",
    "hi_green": "\033[0;92m",
    "hi_yellow": "\033[0;93m",
    "hi_blue": "\033[0;94m",
    "hi_purple": "\033[0;95m",
    "hi_cyan": "\033[0;96m",
    "hi_white": "\033[0;97m",

    # Bold High Intensity
    "bold_hi_black": "\033[1;90m",
    "bold_hi_red": "\033[1;91m",
    "bold_hi_green": "\033[1;92m",
    "bold_hi_yellow": "\033[1;93m",
    "bold_hi_blue": "\033[1;94m",
    "bold_hi_purple": "\033[1;95m",
    "bold_hi_cyan": "\033[1;96m",
    "bold_hi_white": "\033[1;97m",

    # High Intensity backgrounds
    "bg_hi_black": "\033[0;100m",
    "bg_hi_red": "\033[0;101m",
    "bg_hi_green": "\033[0;102m",
    "bg_hi_yellow": "\033[0;103m",
    "bg_hi_blue": "\033[0;104m",
    "bg_hi_purple": "\033[0;105m",
    "bg_hi_cyan": "\033[0;106m",
    "bg_hi_white": "\033[0;107m",
}
Colors = Literal[
    "black", 
    "red", 
    "green", 
    "yellow", 
    "blue", 
    "purple", 
    "cyan", 
    "white", 

    "reset", 
    "italic", 
    "bold_italic", 
    "underline", 
    "strikethrough"

    "bold_black", 
    "bold_red", 
    "bold_green", 
    "bold_yellow", 
    "bold_blue", 
    "bold_purple", 
    "bold_cyan", 
    "bold_white", 

    "underline_black", 
    "underline_red", 
    "underline_green", 
    "underline_yellow", 
    "underline_blue", 
    "underline_purple", 
    "underline_cyan", 
    "underline_white", 

    "bg_black", 
    "bg_red", 
    "bg_green", 
    "bg_yellow", 
    "bg_blue", 
    "bg_purple", 
    "bg_cyan", 
    "bg_white", 

    "hi_black", 
    "hi_red", 
    "hi_green", 
    "hi_yellow", 
    "hi_blue", 
    "hi_purple", 
    "hi_cyan", 
    "hi_white", 

    "bold_hi_black", 
    "bold_hi_red", 
    "bold_hi_green", 
    "bold_hi_yellow", 
    "bold_hi_blue", 
    "bold_hi_purple", 
    "bold_hi_cyan", 
    "bold_hi_white", 

    "bg_hi_black", 
    "bg_hi_red", 
    "bg_hi_green", 
    "bg_hi_yellow", 
    "bg_hi_blue", 
    "bg_hi_purple", 
    "bg_hi_cyan", 
    "bg_hi_white", 
]


def print_separator(text: str = None, sep_type: SepType = "NORMAL", color: Colors = None, out_file: TextIO = None) -> None:
    """Prints a text with a line that separes the bash outputs. The size of this line is controled by sep_type

    Args:
        text (str): Text to print.
        sep_type (Literal['SHORT', 'NORMAL', 'LONG', 'SUPER', 'START'], optional): Type of the separation line. Defaults to "NORMAL".
        out_file (Optional[TextIO], optional): Print the line somewere that's not the std bash (keed None for bash). Defaults to None.
    """

    sep = _separators.get(sep_type.lower(), "") # If the separator is not there do it with ''
    if not sep:
        print_warn("WARNING: No separator with that label", out_file=out_file)

    if sep_type == "SUPER":
        print_color(sep, color=color, out_file=out_file)
        if text:
            print_color(f"{text:^{len(sep)}}", color=color, out_file=out_file)
        print_color(sep + "\n", color=color, out_file=out_file)
    elif sep_type == "START":
        print_color(sep + "\n", color=color if color else "blue", out_file=out_file)
        if text:
            print_color(f"{text:^{len(sep)}}\n", color=color if color else "blue", out_file=out_file)
        print_color(sep + "\n", color=color if color else "blue", out_file=out_file)
    else:
        print_color(sep, color=color, out_file=out_file)
        if text:
            print_color(f"{text:^{len(sep)}}\n", color=color, out_file=out_file)


def print_color(text: str, color: Colors = "reset", log_level: LogLevel = 'info', print_text: bool = True, out_file: TextIO = None) -> str:
    """Prints the text with a certain color

    Args:
        text (str): Text to print
        color (Literal['red', 'green', 'blue', 'reset'...], optional): Color to use. Defaults to "reset".
        print_text bool: Whether or not to print the color text (if false it will return it)
        out_file (Optional[TextIO], optional): Print the line somewere that's not the std bash (keed None for bash). Defaults to None.

    Return: 
        str: Text with colors
    """
    color =  _colors.get(color, _colors['reset'])
    text: str = f"{color}{text}{_colors['reset']}"

    if print_text:
        print_log(f"{text}", log_level=log_level, out_file=out_file)

    return text


def print_warn(text: str, color: Colors = "yellow", prefix: str = '', suffix: str = '', out_file: TextIO = None) -> str:
    """Format and print a warning message surrounded by ⚠️ emojis.

    Args:
        text (str): The message to display as a warning.
        color (Colors, optional): The color of the warning text. Defaults to "yellow".
        prefix (str, optional): Text to prepend before the warning. Defaults to ''.
        suffix (str, optional): Text to append after the warning. Defaults to ''.
        out_file (Optional[TextIO], optional): Print the line somewere that's not the std bash (keed None for bash). Defaults to None.

    Returns:
        str: The formatted warning text with color and emojis.
    """
    return print_color(f"{prefix}⚠️{text}⚠️{suffix}", color=color, log_level="warning", out_file=out_file)

def print_error(text: str, color: Colors = "red", prefix: str = '', suffix: str = '', out_file: TextIO = None) -> str:
    """Format and print an error message surrounded by ❌ emojis.

    Args:
        text (str): The message to display as an error.
        color (Colors, optional): The color of the error text. Defaults to "red".
        prefix (str, optional): Text to prepend before the error. Defaults to ''.
        suffix (str, optional): Text to append after the error. Defaults to ''.
        out_file (Optional[TextIO], optional): Print the line somewere that's not the std bash (keed None for bash). Defaults to None.

    Returns:
        str: The formatted error text with color and emojis.
    """
    return print_color(f"{prefix}❌{text}❌{suffix}", color=color, log_level="error", out_file=out_file)


# ==========================================================================================
#                                    CLEAR LINES
# ==========================================================================================
def print_status(msg: str, log_level: LogLevel = None, out_file: TextIO = None):
    """Prints a dynamic status message on the same terminal line.

    Useful for updating progress or status in-place (e.g. during loops),
    preventing multiple lines of output.

    Args:
        msg (str): Message to display.
        log_level (Optional[LogLevel], optional): Log level to use. Defaults to None.
        out_file (Optional[TextIO], optional): To print the line somewere that's not the std bash (keed None for bash). Defaults to None.
    """
    if log_level is None:
        log_level = get_base_log_level()
    clear_line = " " * _separators_max_length  # assume max 120 chars per line
    print_log(f"{clear_line}\r{msg}\r", end="\r", log_level=log_level, out_file=out_file)

def clear_status(log_level: LogLevel = None, out_file: TextIO = None):
    """Clears the previous status line
    Args:
        log_level (Optional[LogLevel], optional): Log level to use. Defaults to None.
        out_file (Optional[TextIO], optional): To print the line somewere that's not the std bash (keed None for bash). Defaults to None.
    """
    if log_level is None:
        log_level = get_base_log_level()
    print_status("", log_level=log_level, out_file=out_file)

def clear_bash(n_lines: int = 1, out_file: TextIO = None) -> None:
    """Cleans the bash output by removing the last n lines.

    Args:
        n_lines (int, optional): Number of lines to remove. Defaults to 1.
        out_file (Optional[TextIO], optional): To print the line somewere that's not the std bash (keed None for bash). Defaults to None.
    """
    print_log("\033[F\033[K"*n_lines, end="", out_file=out_file)  # Move cursor up one line and clear that line

def print_clear_bash(text: str, n_lines: int = 1, log_level: LogLevel = None, out_file: TextIO = None) -> None:
    """Cleans the bash output by removing the last n lines.

    Args:
        n_lines (int, optional): Number of lines to remove. Defaults to 1.
        out_file (Optional[TextIO], optional): To print the line somewere that's not the std bash (keed None for bash). Defaults to None.

    """
    if log_level is None:
        log_level = get_base_log_level()
    clear_bash(n_lines, out_file=out_file)
    print_log(text, log_level=log_level, out_file=out_file)


def print_utf_8(text: str, print_text: bool = True, out_file: TextIO = None) -> str:
    """Decode escaped Unicode sequences in a string and optionally print it.

    Encodes the input string to UTF-8, decodes escape sequences (e.g. "\\u00e9"),
    replaces "\\n" with newlines, and returns the processed text.

    Args:
        text (str): Input text containing escaped Unicode characters.
        print_text (bool, optional): If True, print the processed text. Defaults to True.
        out_file (Optional[TextIO], optional): To print the line somewere that's not the std bash (keed None for bash). Defaults to None.

    Returns:
        str: The decoded and formatted text.
    """
    text = text.encode("utf-8").decode("unicode_escape").replace("\\n", "\n")
    if print_text:
        print_log(text, out_file=out_file)
    return text

# ==========================================================================================
#                                    TIME FORMATTING
# ==========================================================================================
def print_time(sec: float, n_files: Optional[int] = None, space: bool = False, prefix: str = "", sufix: str = "", out_file: Optional[TextIO] = None) -> None:
    """Given a certain number of seconds, parse it to Formatted time string (e.g., '01 hrs, 05 mins, 30.1234 sec').
    If not enough seconds for hours, just '05 mins, 30.1234 sec'.
    If neither enogh seconds for minuts, just parse '30.1234 sec'.
    Optionally you can add a 'number of files' to get avg. metrics as well as extra config for better printing.
    Optionally you can pass a file to print everyting there.

    Args:
        sec (float): Number of seconds
        n_files (Optional[int], optional): Number of files to add an avg. Defaults to None.
        space (bool, optional): To add a space before the print. Defaults to False.
        prefix (str, optional): To add a prefix before the print. Defaults to "".
        sufix (str, optional): To add a sufix after the print. Defaults to "".
        out_file (Optional[TextIO], optional): Print the line somewere that's not the std bash (keed None for bash). Defaults to None.
    """
    if space:
        print_log("")
    
    if n_files is not None:
        message = f"{n_files:4} files in: {parse_seconds_to_minutes(sec)}."
        message += f" Per document: {parse_seconds_to_minutes(sec / n_files)}"
    else:
        message = f"Time: {parse_seconds_to_minutes(sec)}"

    message = f"{prefix}{message}{sufix}."
    
    if out_file:
        print(message, file=out_file)
    else:
        print_log(message)
